;; Copyright 2006, 2007, 2009, 2010, 2012, 2013, 2015 Kevin Ryde.
;;
;; This file is part of accjournal.
;;
;; accjournal is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; accjournal is distributed in the hope that it will be useful, but WITHOUT
;; ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
;; FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
;; more details.
;;
;; You should have received a copy of the GNU General Public License along
;; with accjournal; see the file COPYING.  Failing that, go to
;; <http://www.gnu.org/licenses/>.


;; f5 all
;; C-c C-f format one

;; C-c c-a maybe align
;; C-c c-p preview, but maybe print
;; C-c c-v too close to C-x C-c
;; C-c c-o open variant

;; C-c c-t account "t"
;; C-c c-l listing all


;;-----------------------------------------------------------------------------
;; compile
;; emacs20 dolist

(progn
  (add-to-list 'load-path (expand-file-name "../el/devel"))
  (require 'my-byte-compile)
  (my-byte-compile))
(progn
  (add-to-list 'load-path (expand-file-name "."))
  (require 'my-byte-compile)
  (my-show-autoloads))


;;-----------------------------------------------------------------------------
;; *?
(let ((str "aax xxx "))
  (string-match "\\([^@\t]+?\\)[ \t]*\\(|.*\\)?$" str)
  (match-string 1 str))

;;-----------------------------------------------------------------------------
;; integer overflow

most-positive-fixnum
(/ (log most-positive-fixnum) (log 2))
(/ (log most-positive-fixnum) (log 10))
(expt 2 29)
(1- (expt 2.0 29))

(floatp (string-to-number "1000001729688"))

(accjournal-add most-positive-fixnum 0)
(accjournal-add most-positive-fixnum 1)
(accjournal-add most-positive-fixnum -1)
(accjournal-add most-positive-fixnum most-positive-fixnum)

(accjournal-add most-negative-fixnum 0)
(accjournal-add most-negative-fixnum 1)
(accjournal-add most-negative-fixnum -1)
(accjournal-add most-negative-fixnum most-negative-fixnum)

(let ((lst (make-list 6 'x)))
  (setf (nth 3 lst) 'z)
  lst)
(let ((lst (make-list 6 'x)))
  (setcar (nthcdr 3 lst) 'z)
  lst)


;;-----------------------------------------------------------------------------

(defvar accjournal-minibuffer-local-completion-map
  (let ((m (make-sparse-keymap)))
    (define-key m " " 'self-insert-command)
    m)
  "Inherited `minibuffer-local-completion-map' but with space self-inserting.")

  (set-keymap-parent accjournal-minibuffer-local-completion-map
                     minibuffer-local-completion-map)

(minibuffer-local-completion-map
         accjournal-minibuffer-local-completion-map)
        

;;-----------------------------------------------------------------------------

(or (progn
      (goto-char (window-start))
      (forward-line -1)
      (goto-char (1- (point)))
      (search-forward accjournal-separator nil t))
    (progn
      (goto-char (window-end))
      (search-backward accjournal-separator nil t))
    (goto-char (point-min)))


(let ((line      (count-lines (point-min) (point)))
      (column    (current-column))
      (name      (accjournal-output-account-name))
      (name-line (count-lines (window-start) (point))))
  )

(list window
      point-line
      point-column
      start-line
      start-column
      name
      name-line
                         (count-lines (point-min) (window-point window))
                         (save-excursion
                           (goto-char (window-point window))
                           (current-column))
                         (count-lines (point-min) (window-start window))
                         (save-excursion
                           (goto-char (window-start window))
                           (current-column)))


(defmacro accjournal-save-point-and-windows (buffer &rest body)
  (declare (debug t))
  `(let ((accjournal-save-point-and-windows--buffer ,buffer)
         accjournal-save-point-and-windows--wlist)
     (dolist (frame (frame-list))
       (dolist (window (window-list frame))
         (when (eq buffer (window-buffer window))
           (push (list window
                       (count-lines (point-min) (point))
                       (current-column)
                       (count-lines (point-min) (window-start window))
                       (save-excursion
                         (goto-char (window-start window))
                         (current-column)))
                 accjournal-save-point-and-windows--wlist))))
     (save-excursion
       )))

;;------------------------------------------------------------------------------
;; loop

(cl-prettyexpand '(loop for window being the windows))
(block nil
  (let* ((window (selected-window))
         (--cl-var-- nil))
    (while (prog1
               (not (eq window --cl-var--))
             (or --cl-var-- (setq --cl-var-- window)))
      (setq window (next-window window)))
    nil))

(cl-prettyexpand '(loop for i from 1 to 10 collect i))

(block nil
  (let* ((i 1)
         (--cl-var-- nil))
    (while (<= i 10)
      (setq --cl-var-- (cons i --cl-var--))
      (setq i (+ i 1)))
    (nreverse --cl-var--)))



(progn
  (setq align-rules-list nil)
)


;;-----------------------------------------------------------------------------
;; align

30 Jun 85  20000.00  Bank     -> !InvHouse     | buy
30 Jun 85  20000.00  Bank     -> !InvHouse
30 May 86   5000.00  Bank     -> InvHouse      | renovation
30 Jun 86    800.00  InvHouse -> 85/86 deducts | depreciation

$ 20000.00
$ 5000.00
$ 800.00


;; (progn
;;   (setq align-rules-list nil)
;;   (push `(accjournal-note
;;           (regexp . ,(concat accjournal-date-regexp
;;                              "[ \t]+"
;;                              accjournal-amount-regexp
;;                              "[ \t]+.*?"  ;; from
;;                              "[ \t]+->"   ;;
;;                              "[ \t]+.*?"  ;; to
;;                              "\\([ \t]+\\)|"))  ;; note
;;           (modes    . '(emacs-lisp-mode accjournal-mode)))
;;         align-rules-list)
;;   (push `(accjournal-fields
;;           (regexp . ,(concat accjournal-date-regexp
;;                              "[ \t]+"
;;                              accjournal-amount-regexp
;;                              "\\([ \t]+\\).*?"  ;; from
;;                              "\\([ \t]+\\)->"   ;;
;;                              "\\([ \t]+\\)."))    ;; to
;;           (group  . (1 2 3))
;;           (modes    . '(emacs-lisp-mode accjournal-mode))
;;           )
;;         align-rules-list)
;;   (push `(accjournal-amount
;;           (regexp   . ,(concat  accjournal-date-regexp
;;                                 " \\( *" accjournal-amount-regexp "\\)"))
;;           (justify  . t)
;;           (tab-stop . nil)
;;           (modes    . '(emacs-lisp-mode accjournal-mode))
;;           )
;;         align-rules-list))

'
'(add-to-list 'align-rules-list
              `(accjournal-notebar
                (regexp   . ,align-re)))

;;   (add-to-list 'align-rules-list
;;                `(accjournal-to    (regexp . ,align-re)
;;                                   (group  . (3))))
;;   (add-to-list 'align-rules-list
;;                `(accjournal-arrow (regexp . ,align-re)
;;                                   (group  . (2))))
;;   (add-to-list 'align-rules-list
;;                `(accjournal-from  (regexp . ,align-re)
;;                                   (group  . (1))))
;;                ;;(regexp   . "->")
;;                ))

;;   (add-to-list 'align-rules-list
;;                `(accjournal-arrow
;;                  (regexp   . ,(concat accjournal-date-regexp
;;                                       "[ \t]+"
;;                                       accjournal-amount-regexp
;;                                       "[ \t]+.*?"  ;; from
;;                                       "\\([ \t]+\\)->"   ;;
;;                                       ))))
;;   (add-to-list 'align-rules-list
;;                `(accjournal-arrow
;;                  (regexp   . ,
;;                  ;; (groups   . (1 2 3 4))
;;                  ;; (modes    . '(emacs-lisp-mode accjournal-mode))
;;                  )))


;;-----------------------------------------------------------------------------

(defcustom accjournal-decimals 2
  ""
  :type  'integer
  :group 'accjournal)
;;;###autoload
(put 'accjournal-decimals 'safe-local-variable 'integerp)


;;-----------------------------------------------------------------------------

;;(setq debug-on-error t)
;;(accjournal-make)

;; (progn (kill-buffer (get-buffer-create "*Compile-Log*")) (byte-compile-file "journal.el") (load "/home/gg/tax/journal.elc"))

(defun accjournal-mmm< (x y)
  (< (accjournal-mmm-to-month x)
     (accjournal-mmm-to-month y)))
;; (accjournal-mmm< "Jun" "Aug")

;; eg " 1 Jul 04"
(defun accjournal-date< (x y)
  (let ((x-year (accjournal-yy-to-yyyy (substring x 7 9)))
        (y-year (accjournal-yy-to-yyyy (substring y 7 9))))
    (or (string< x-year y-year)
        (and (string= x-year y-year)

             (let ((x-month (substring x 3 6))
                   (y-month (substring y 3 6)))
               (or (accjournal-mmm< x-month y-month)
                   (and (string= x-month y-month)

                        (< (string-to-number (substring x 0 2))
                           (string-to-number (substring y 0 2))))))))))
;; (accjournal-date< "30 Jun 04" "23 Aug 04")
;; (accjournal-date< "30 Jun 04" "23 Aug 04")

(defun accjournal-date-max (x y)
  (if (accjournal-date< x y)
      y x))
;; (accjournal-date-max " 1 Jan 00" " 1 Jan 01")
;; (accjournal-date-max " 1 Jan 99" " 1 Jan 00")

      (if (get-buffer "*accjournal*")
          (kill-buffer "*accjournal*"))
      (switch-to-buffer-other-window "*accjournal*")

      (let ((account (assoc-default name accjournal-accounts-alist)))
        (if account
            (insert (format "%s\n%s\n\n%s\n"
                            name
                            (apply 'concat (reverse
                                            (aref account
                                             accjournal-acc-translist)))
                            (if (aref account accjournal-acc-closedp)
                                "Closed" "(Open)")))
          (insert (format "Not found: %s\n\n" account))))

      (goto-char (point-min))

(defun accjournal-yy-to-yyyy (str)
  "Convert STR like \"95\" to integer 1995, or \"2010\" to 2010."
  (if (numberp (compare-strings str 0 1 "9" 0 1))
      (concat "20" str)
      (concat "19" str)))
;;(accjournal-yy-to-yyyy "00")



;;-----------------------------------------------------------------------------

(defun accjournal-string-trim-white (str)
  "Return STR without leading or trailing whitespace characters."
  (if (string-match "\\`[ \t\n]+" str)
      (setq str (replace-match "" t t str)))
  (if (string-match "[ \t\n]+\\'" str)
      (setq str (replace-match "" t t str)))
  str)
            (setq from-name (accjournal-string-trim-white from-name))
            (setq to-name   (accjournal-string-trim-white to-name))

    ;;-------------------------------------------------------------------------
    ;; accjournal-string-trim-white

    (mapcar (lambda (elem)
              (let* ((str (nth 0 elem))
                     (want (nth 1 elem))
                     (got (accjournal-string-trim-white str)))
                (unless (equal got want)
                  (error "`accjournal-string-trim-white' str %S got %S want %S"
                         str got want)))
              (setq count (1+ count)))

            '(("  x"     "x")
              ("x  "     "x")
              ("  x  "   "x")
              ("x  y "   "x  y")
              ("  x  y " "x  y")))

;;-----------------------------------------------------------------------------
accjournal-amount-width

  :type  'integer
  :group 'accjournal)
;;;###autoload
(put 'accjournal-amount-width 'safe-local-variable 'integerp)


;;-----------------------------------------------------------------------------
;; open date

          nil   ;; opendateN
(defconst accjournal-acc-opendateN  3)
           (unless (aref from-acc accjournal-acc-opendateN)
             (setf (aref from-acc accjournal-acc-opendateN) date-N))
           (unless (aref to-acc accjournal-acc-opendateN)
             (setf (aref to-acc accjournal-acc-opendateN) date-N))

;;-----------------------------------------------------------------------------
;; how `loop' comes out

(progn
  (defun foo ()
    (let ((h (make-hash-table :test 'equal)))
      (loop for str in '("Jan" "Feb" "Mar" "Apr" "May" "Jun"
                         "Jul" "Aug" "Sep" "Oct" "Nov" "Dec")
            for i from 1
            do
            (puthash str i h)
            (puthash (upcase str) i h)
            (puthash (downcase str) i h))
      h))
  (byte-compile 'foo)
  (disassemble 'foo))

(progn
  (byte-compile 'accjournal-make)
  (disassemble 'accjournal-make))


;;-----------------------------------------------------------------------------
;; speed let vs setq

(disassemble 'my-speed-let)
(disassemble 'my-speed-setq)
(progn
  (defun my-dummy () nil)
  (byte-compile 'my-dummy)

  (defun my-speed-let ()
    (let ((t1 (float-time))
          t2)
      (dotimes (i 1000000)
        (let ((a 1)
              (b 1)
              (c 1)
              (d 1)
              (e 1)
              (f 1)
              (g 1)
              (h 1)
              (i 1)
              (j 1)
              (k 1)
              (l 1)
              (m 1)
              (n 1)
              (o 1))
          (my-dummy)))
      (setq t2 (float-time))
      (- t2 t1)))
  (byte-compile 'my-speed-let)

  (defun my-speed-setq ()
    (let ((t1 (float-time))
          t2)
      (let (a b c d e f g h i j k l m n o)
        (dotimes (z 1000000)
          (setq a 1)
          (setq b 1)
          (setq c 1)
          (setq d 1)
          (setq e 1)
          (setq f 1)
          (setq g 1)
          (setq h 1)
          (setq i 1)
          (setq j 1)
          (setq k 1)
          (setq l 1)
          (setq m 1)
          (setq n 1)
          (setq o 1)
          (my-dummy)))
      (setq t2 (float-time))
      (- t2 t1)))
  (byte-compile 'my-speed-setq))



;;-----------------------------------------------------------------------------
;; accjournal-font-lock-warning

(cl-prettyexpand (symbol-function 'accjournal-font-lock-warning))
(function
 (lambda (limit)
   "Search for a bad accjournal line up to LIMIT.
This function is designed for use as a `font-lock-keywords'
matcher.

Return nil if there's no bad lines.  Return non-nil if there's a
bad line, setting match data 0 to encompass it and leaving point
at the start of the following line (ready for a further search)."
   (beginning-of-line)
   (block nil
     (while (< (point) limit)
       (if (or (looking-at "[ 	]*$") (looking-at comment-start-skip) (looking-at accjournal-line-regexp)) nil (looking-at ".*
?") (goto-char (match-end 0)) (cl-block-throw '--cl-block-nil-- t))
       (forward-line 1))
     nil)))

;;-----------------------------------------------------------------------------
;; define-derived-mode expansion

(progn
  (load (expand-file-name "../accjournal.el"))
  (let ((func (symbol-function 'accjournal-mode)))
    (setq func (copy-list func))
    (setcar (cddr func) "")
    (pp func (current-buffer))))


;;-----------------------------------------------------------------------------
;; auto-complete.el
;; /usr/share/doc/auto-complete-el/doc/manual.txt
;; /usr/share/emacs/site-lisp/auto-complete/auto-complete.el

(progn

  (find-file "d-accjournal.txt")
  (auto-complete-mode 1)
  (set (make-local-variable 'ac-sources)
       '(accjournal-ac-source-accounts))
  )

    ;; (prefix ,(concat "\\(^" accjournal-date-regexp
    ;;                  "[ \t]+" accjournal-amount-regexp
    ;;                  "[ \t]*"
    ;;                  "\\|!\\|->"
    ;;                  "\\)"))


;;-----------------------------------------------------------------------------
;; old bit

;; (defun accjournal-completion-make-lazy-accounts ()
;;   "Return a completion handler for account names in the current buffer.
;; The return is designed for use by `completing-read'."
;;   (lexical-let ((buffer (current-buffer))
;;                 table done)
;;     (lambda (str pred action)
;;       (unless done
;;         (setq done t)
;;         (setq table (with-current-buffer buffer
;;                       (accjournal-completion-accounts))))
;;       (accjournal--complete-with-action action table str pred))))

